/*
			Copyright (c) 1994 by
			Advanced Visual Systems Inc.
			All Rights Reserved

	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.

	This file is under Perforce control
	$Id: //depot/express/fcs992/modules/stream.c#1 $
*/
#include <avs/util.h>
#include <avs/err.h>
#include <avs/om.h>
#include <avs/fld.h>
#include <avs/arr.h>
#include <avs/dv_util.h>
#include <avs/mat.h>

#define ERR_RETURN(A) ERRerror("stream",1,ERR_ORIG, A); return(0);
#define CONN_BLOCK    1024
#define MAX_NAME_SIZE 1024

int FUNCstream (OMobj_id in, OMobj_id probe, int order, int forward, int nseg,
                int max_seg, double min_vel, int stream_comp, int ribbons,
                double rib_width, double rib_angle, OMobj_id out);

int DVstream_update(OMobj_id elem_id)
{
	OMobj_id in, out, elm, probe;
	int type, size;
	float  *point;
	double min_vel;
	int   order, forward, backward, nseg, stream_comp, max_seg, ribbons;
	double rib_width, rib_angle;

	in = OMfind_subobj(elem_id, OMstr_to_name("in"), OM_OBJ_RD);
	out = OMfind_subobj(elem_id, OMstr_to_name("out"), OM_OBJ_RD);

	probe = OMfind_subobj(elem_id, OMstr_to_name("probe"), OM_OBJ_RD);

	elm = OMfind_subobj(elem_id, OMstr_to_name("order"), OM_OBJ_RD);
	OMget_int_val(elm, &order);

	elm = OMfind_subobj(elem_id, OMstr_to_name("forw_back"), OM_OBJ_RD);
	OMget_int_val(elm, &forward);

	elm = OMfind_subobj(elem_id, OMstr_to_name("backward"), OM_OBJ_RD);
	OMget_int_val(elm, &backward);

	elm = OMfind_subobj(elem_id, OMstr_to_name("nseg"), OM_OBJ_RD);
	OMget_int_val(elm, &nseg);

	elm = OMfind_subobj(elem_id, OMstr_to_name("max_seg"), OM_OBJ_RD);
	OMget_int_val(elm, &max_seg);

	elm = OMfind_subobj(elem_id, OMstr_to_name("stream_comp"), OM_OBJ_RD);
	OMget_int_val(elm, &stream_comp);

	elm = OMfind_subobj(elem_id, OMstr_to_name("min_vel"), OM_OBJ_RD);
	OMget_real_val(elm, &min_vel);

	elm = OMfind_subobj(elem_id, OMstr_to_name("ribbons"), OM_OBJ_RD);
	OMget_int_val(elm, &ribbons);
	if (ribbons) {
		elm = OMfind_subobj(elem_id, OMstr_to_name("rib_width"), OM_OBJ_RD);
		OMget_real_val(elm, &rib_width);
		elm = OMfind_subobj(elem_id, OMstr_to_name("rib_angle"), OM_OBJ_RD);
		OMget_real_val(elm, &rib_angle);
	}
	else {
		rib_width = 0;
		rib_angle = 0;
	}
	if (FUNCstream(in, probe, order, forward, nseg, max_seg, min_vel,
			stream_comp, ribbons, rib_width, rib_angle, out)) {
		return(1);
	}
	else return(0);
}


int FUNCstream (OMobj_id in, OMobj_id probe, int order, int forward, int nseg,
                int max_seg, double min_vel, int stream_comp, int ribbons,
                double rib_width, double rib_angle, OMobj_id out)
{
	int   cs, nsets, ncells, nnodes;
	int   set, cell, nspace, stat, nalloc, out_veclen;
	int   ndim, size, veclen, n, nprobes, i;
	float uvw[3], *min_ext, *max_ext, point[3], end_point[3], *probe_xyz, width_end, angle_end;
	float xform[4][4], *xfm, in_xform[4][4], *in_xfm, t_xform[4][4];
	char  *node_data;
	char  *block_table, *mesh_info;
	int   *cells, npoints, probe_nspace, dtype, null_flag, dir;
	float *points, *uvws, *values, *times, *mags, *out_data;
	int   out_nnodes, out_ncells, node_size, *connect, conn_size;
	char  label[MAX_NAME_SIZE], units[MAX_NAME_SIZE], cell_name[MAX_NAME_SIZE];
	double null_value;
	OMobj_id   cell_set;
	int max_points;
	int interrupt;


	/********************************/
	/*   Free pre-allocated arrays  */
	/********************************/
	if (FLDset_ncell_sets (out, 0) != 1) {
		ERR_RETURN("Error setting ncell_sets");
	}
	out_data = points = NULL;
	if (FLDset_nnodes (out, 0) != 1) {
		ERR_RETURN("Error setting nnodes");
	}
	if (FLDset_coord(out, points, 0, OM_SET_ARRAY_FREE) != 1) {
		ERR_RETURN("Error setting coord");
	}
	if (FLDset_node_data_ncomp (out, 0) != 1) {
		ERR_RETURN("Error setting nnode_data");
	}

	/**  Check input **/

	if (FLDget_nnodes(in, &nnodes) != 1) {
		ERR_RETURN("cannot get nnodes");
	}
	if (FLDget_nspace(in, &nspace) != 1) {
		ERR_RETURN("cannot get nnodes");
	}
	if (FLDget_node_data_veclen(in, stream_comp, &veclen) != 1) {
		ERR_RETURN("Error getting veclen");
	}
	if (veclen < 2) {
		ERR_RETURN("cannot create streamlines for scalar");
	}

	stat = FLDget_mesh_info(in, &mesh_info);
	if (!stat) {
		ERR_RETURN("cannot create cell table");
	}

	stat = FLDinterp_init(in, mesh_info, &block_table);
	if (stat != 1) {
		FLDfree_mesh_info(mesh_info);
		ERR_RETURN("no block table created");
	}

	if (FLDget_node_data(in, stream_comp, &dtype, &node_data, &size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("cannot get node data");
	}
	if (FLDget_node_null_data(in, stream_comp, &null_flag, (char *)&null_value) != 1) {
		ERR_RETURN("cannot get null data");
	}

	if (FLDget_nnodes(probe, &nprobes) != 1) {
		ERR_RETURN("cannot get nnodes for probe");
	}
	if (FLDget_nspace(probe, &probe_nspace) != 1) {
		ERR_RETURN("cannot get nspace for probe");
	}
	if (FLDget_coord(probe, &probe_xyz, &size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("cannot get coordinates for probe");
	}
	stat = FLDget_xform(probe, (float *)xform);
	if (stat < 0) {
		ERR_RETURN("cannot get xform for probe");
	}
	else if (stat == 0) {
		xfm = (float *)0;
	}
	else if (MATmat_is_identity((float *)xform, 4))
		xfm = (float *)0;
	else
		xfm = (float *)xform;
	stat = FLDget_xform(in, (float *)in_xform);
	if (stat < 0) {
		ERR_RETURN("cannot get xform for field");
	}
	else if (stat == 0) {
		in_xfm = (float *)0;
	}
	else if (MATmat_is_identity((float *)in_xform, 4))
		in_xfm = (float *)0;
	else
		in_xfm = (float *)in_xform;
	if (in_xfm) {
		MATmat4_inverse(t_xform, (Matr4 *)in_xfm);
		if (xfm)
			MATmat4_multiply((Matr4 *)xfm, t_xform, (Matr4 *)xfm);
		else
			xfm = (float *)t_xform;
	}
	if (forward)
		dir = 1;
	else
		dir = -1;

	max_points = max_seg+1;
	out_nnodes = 0;
	out_ncells = 0;
	if (ribbons) {
		strcpy(cell_name, "Quad");
		nalloc = 2;
		conn_size = 4*(max_points-1);
		node_size = 2*max_points;
		out_veclen = 1;
		uvws = (float *)malloc(3*max_points*sizeof(float));
		times = (float *)malloc(max_points*sizeof(float));
		mags = (float *)malloc(max_points*sizeof(float));
		cells = (int *)malloc(max_points*sizeof(int));
		if (uvws == NULL || times == NULL || mags == NULL || cells == NULL) {
			ERR_RETURN("cannot allocate data");
		}
	}
	else {
		strcpy(cell_name, "Polyline");
		nalloc = 1;
		conn_size = CONN_BLOCK;
		node_size = max_points;
		out_veclen = veclen;
		uvws = (float *)0;
		times = (float *)0;
		mags = (float *)0;
		cells = (int *)0;
	}
	connect = (int *)ARRalloc(NULL, DTYPE_INT, conn_size, NULL);
	points = (float *)ARRalloc(NULL, DTYPE_FLOAT, nspace*node_size, NULL);
	out_data = (float *)ARRalloc(NULL, DTYPE_FLOAT, out_veclen*node_size, NULL);
	values = (float *)malloc(veclen*max_points*sizeof(float));
	if (connect==NULL || points == NULL || out_data==NULL || values == NULL) {
		ERR_RETURN("cannot allocate data");
	}

	for (n=0; n<nprobes; n++) {
		if (n & 0x1) {
			OMstatus_check(100*n/nprobes,"streamlines_update",&interrupt);
			if (interrupt) {
				if (values)
					free(values);
				if (uvws)
					free(uvws);
				if (mags)
					free(mags);
				if (times)
					free(times);
				if (cells)
					free(cells);
				ARRfree(connect);
				ARRfree(points);
				ARRfree(out_data);
				ARRfree(node_data);
				ARRfree(probe_xyz);
				FLDfree_mesh_info(mesh_info);
				FLDinterp_end(block_table);
				ERR_RETURN("streamlines are not created");
			}
		}
		width_end = 0;
		angle_end = 0;
		point[0] = point[1] = point[2] = 0.0;
		memcpy(point, probe_xyz+n*probe_nspace, probe_nspace*sizeof(float));
		if (xfm)
			MATvec3_mat4_multiply(point, (Matr4 *)xfm);
		if (node_size < out_nnodes+nalloc*max_points) {
			node_size += nalloc*max_points;
			points = (float *)ARRrealloc(points, DTYPE_FLOAT,
						     nspace*node_size, NULL);
			out_data = (float *)ARRrealloc(out_data, DTYPE_FLOAT,
						       out_veclen*node_size, NULL);
			if (points == NULL || out_data == NULL) {
				ERR_RETURN("cannot allocate data");
			}
		}
		UTILstream_line(mesh_info, block_table,
				node_data, dtype, veclen, null_flag, (char *)&null_value,
				order, dir, nseg, min_vel, point,
				max_points, &npoints, points+out_nnodes*nspace, values,
				ribbons, cells, uvws, mags, times, end_point);
		if (npoints>1) {
			if (ribbons) {
				stat = UTILribbon(mesh_info, npoints, nspace,
						  points+out_nnodes*nspace, values,
						  cells, uvws, mags, times,
						  veclen, node_data, dtype,
						  null_flag, (char *)&null_value,
						  rib_width, 3.14*rib_angle/180.0, &width_end, &angle_end);
				if (!stat) {
					ERR_RETURN("Error creating ribbons");
				}
				if (conn_size <= 4*(out_ncells+npoints-1)) {
					conn_size += 4*(max_points-1);
					connect = (int *)ARRrealloc(connect, DTYPE_INT,
								    conn_size, NULL);
					if (connect == NULL) {
						ERR_RETURN("cannot allocate connectivity");
					}
				}
				    for (i=0; i<npoints-1; i++) {
					connect[out_ncells*4] = out_nnodes+i;
					connect[out_ncells*4+1] = out_nnodes+1+i;
					connect[out_ncells*4+2] = out_nnodes+1+npoints+i;
					connect[out_ncells*4+3] = out_nnodes+npoints+i;
					out_ncells++;
				}
				memcpy(out_data+out_veclen*out_nnodes, mags, npoints*out_veclen*sizeof(float));
				memcpy(out_data+out_veclen*(out_nnodes+npoints), mags, npoints*out_veclen*sizeof(float));
				out_nnodes += 2*npoints;
			}
			else {
				if (conn_size <= 2*out_ncells) {
					conn_size += CONN_BLOCK;
					connect = (int *)ARRrealloc(connect, DTYPE_INT,
								    conn_size, NULL);
					if (connect == NULL) {
						ERR_RETURN("cannot allocate connectivity");
					}
				}
				memcpy(out_data+out_veclen*out_nnodes, values, npoints*out_veclen*sizeof(float));
				connect[out_ncells*2] = out_nnodes;
				out_nnodes += npoints;
				connect[out_ncells*2+1] = out_nnodes-1;
				out_ncells++;
			}
		}
	}

	if (FLDget_coord_units(in, units, MAX_NAME_SIZE) == 1) {
		if (FLDset_coord_units (out, units) != 1) {
			ERR_RETURN("Error setting units");
		}
	}
	if (FLDset_nspace (out, nspace) != 1) {
		ERR_RETURN("Error setting nspace");
	}
	if (FLDset_nnodes (out, out_nnodes) != 1) {
		ERR_RETURN("Error setting nnodes");
	}

	/****  PUT VELOCITY into DATA  ***/

	if (FLDset_node_data_ncomp (out, 1) != 1) {
		ERR_RETURN("Error setting nnode_data");
	}
	if (FLDget_node_data_units(in, stream_comp, units, MAX_NAME_SIZE) != 1) {
		strcpy(units, "");
	}
	if (FLDget_node_data_label(in, stream_comp, label, MAX_NAME_SIZE) != 1) {
		strcpy(label, "");
	}
	if (FLDset_node_data_comp (out, 0, out_veclen, label, units) != 1) {
		ERR_RETURN("Error setting node component");
	}
	if (out_nnodes) {
		if (FLDset_node_data(out, 0, (char *)out_data, DTYPE_FLOAT,
				     out_veclen*out_nnodes, OM_SET_ARRAY_FREE) != 1) {
			ERR_RETURN("Error setting node data");
		}
		if (FLDcopy_node_minmax(in, out, stream_comp, 0) != 1) {
			ERR_RETURN("Error copying node minmax");
		}
		if (!ribbons) {
			if (FLDcopy_node_minmax_vec(in, out, stream_comp, 0) != 1) {
				ERR_RETURN("Error copying node minmax");
			}
		}
	}

	if (FLDadd_cell_set(out, cell_name) != 1) {
		ERR_RETURN("Error setting cell type");
	}

	if (FLDget_cell_set(out, 0, &cell_set) != 1) {
		ERR_RETURN("Error getting cell set");
	}

	if (ribbons) {
		if (FLDset_ncells(cell_set, out_ncells) != 1) {
			ERR_RETURN("Error setting ncells");
		}
	}
	else {
		if (FLDset_npolys(cell_set, out_ncells) != 1) {
			ERR_RETURN("Error setting ncells");
		}
	}
	if (out_nnodes)
		if (FLDset_coord(out, points, nspace*out_nnodes, OM_SET_ARRAY_FREE) != 1) {
			ERR_RETURN("Error setting coord");
		}
	if (out_ncells) {
		if (ribbons) {
			if (FLDset_node_connect(cell_set, connect, 2*nalloc*out_ncells,
						OM_SET_ARRAY_FREE) != 1) {
				ERR_RETURN("Error setting cell connect list");
			}
		}
		else {
			if (FLDset_poly_connect(cell_set, connect, 2*nalloc*out_ncells,
						OM_SET_ARRAY_FREE) != 1) {
				ERR_RETURN("Error setting cell connect list");
			}
		}
	}
	if (values)
		free(values);
	if (uvws)
		free(uvws);
	if (mags)
		free(mags);
	if (times)
		free(times);
	if (cells)
		free(cells);
	FLDfree_mesh_info(mesh_info);
	FLDinterp_end(block_table);
	ARRfree(node_data);
	ARRfree(probe_xyz);
	return(1);
}

